@@ -8,6 +8,7 @@ global script variables:  i.e. debug=True (python expression)
 
 import email
 import inspect
+import logging
 import os
 import pty
 import resource
@@ -21,58 +22,72 @@ import termios
 import time
 import traceback
 from enum import Enum
+from functools import partial
 from path import Path
 from syslog import syslog
 
-"-flags -f --flag -o=foo --option4=bar param1 param2 ..."
-
 """
-(help, kind, abbrev, type, choices, metavar)
+(help, kind, abbrev, type, choices, usage_name)
 
   - help --> the help message
 
   - kind --> what kind of parameter
     - flag       --> simple boolean
     - option     --> option_name value
-    - keyword    --> key=value syntax (no dashes, key can be any
-                     valid Python identifier)
-    - required   --> just like it says (default)
+    - multi      --> option_name value option_name value
+    - required   --> required_name value
 
   - abbrev is a one-character string (defaults to first letter of
     argument)
 
   - type is a callable that converts the arguments to any Python
-    type; by default there is no conversion and type=None 
+    type; by default there is no conversion and type is effectively str
 
   - choices is a discrete sequence of values used to restrict the
     number of the valid options; by default there are no restrictions
     (i.e. choices=None)
 
-  - metavar is used as the name of the parameter in the help message
+  - usage_name is used as the name of the parameter in the help message
 """
 
-# TODO - understand this
-# metavar has two meanings. For a required argument it is used to change the
-# argument name in the usage message (and only there). By default the metavar is
-# None and the name in the usage message is the same as the argument name. For an
-# option the metavar is used differently in the usage message, which has now the
-# form [--option-name METAVAR]. If the metavar is None, then it is equal to the
-# uppercased name of the argument, unless the argument has a default: then it is
-# equal to the stringified form of the default.
-
 # data
 __all__ = (
-    'Command', 'Script', 'Run',
+    'Command', 'Script', 'Run', 'Spec',
     'InputFile', 'Bool',
-    'FLAG', 'OPTION', 'KEYWORD', 'REQUIRED',
+    'FLAG', 'KEYWORD', 'OPTION', 'MULTI', 'REQUIRED',
     'ScriptionError',
     )
 
+version = 0, 50, 7
+
 try:
     bytes
 except NameError:
     bytes = str
 
+class NullHandler(logging.Handler):
+    """
+    This handler does nothing. It's intended to be used to avoid the
+    "No handlers could be found for logger XXX" one-off warning. This is
+    important for library code, which may contain code to log events. If a user
+    of the library does not configure logging, the one-off warning might be
+    produced; to avoid this, the library developer simply needs to instantiate
+    a NullHandler and add it to the top-level logger of the library module or
+    package.
+    
+    Taken from 2.7 lib.
+    """
+    def handle(self, record):
+        """Stub."""
+
+    def emit(self, record):
+        """Stub."""
+
+    def createLock(self):
+        self.lock = None
+
+logger = logging.getLogger('scription')
+logger.addHandler(NullHandler())
 
 class DocEnum(Enum):
     """compares equal to all cased versions of its name
@@ -108,13 +123,19 @@ class DocEnum(Enum):
     def __ne__(self, other):
         return not self == other
 
+    @classmethod
+    def export_to(cls, namespace):
+        namespace.update(cls.__members__)
+
 
 class SpecKind(DocEnum):
-    FLAG = "True/False setting"
-    OPTION = "variable setting"
-    KEYWORD = "global setting (useful for @Command scripts)"
-    REQUIRED = "required setting"
-globals().update(SpecKind.__members__)
+    REQUIRED = "required value"
+    OPTION = "single value per name"
+    MULTI = "multiple values per name (list form)"
+    FLAG = "boolean value per name"
+    KEYWORD = 'unknown options'
+
+SpecKind.export_to(globals())
 
 class ExecutionError(Exception):
     "errors raised by Execute"
@@ -245,12 +266,12 @@ def log_exception(tb=None):
         exc, err, tb = sys.exc_info()
         lines = traceback.format_list(traceback.extract_tb(tb))
         lines.append('%s: %s\n' % (exc.__name__, err))
-        syslog('Traceback (most recent call last):')
+        logger.critical('Traceback (most recent call last):')
     else:
         lines = tb.split('\\n')
     for line in lines:
         for ln in line.rstrip().split('\n'):
-            syslog(ln)
+            logger.critical(ln)
     return lines
 
 def mail(server, port, message):
@@ -308,16 +329,16 @@ class ScriptionError(Exception):
 class Spec(tuple):
     """tuple with named attributes for representing a command-line paramter
 
-    help, kind, abbrev, type, choices, metavar
+    help, kind, abbrev, type, choices, usage_name
     """
 
     __slots__= ()
-    def __new__(cls, *args):
-        if not args or isinstance(args[0], (str, unicode)):
-            pass
+    def __new__(cls, help=None, kind=None, abbrev=None, type=None, choices=None, usage=None, remove=None):
+        if isinstance(help, tuple):
+            args = help
         else:
-            args = args[0]
-        args = list(args) + [None] * (6 - len(args))
+            args = (help, kind, abbrev, type, choices, usage, remove)
+        args = list(args) + [None] * (7 - len(args))
         if not args[0]:
             args[0] = ''
         if not args[1]:
@@ -343,8 +364,21 @@ class Spec(tuple):
     def choices(self):
         return self[4]
     @property
-    def metavar(self):
+    def usage_name(self):
         return self[5]
+    @property
+    def remove(self):
+        return self[6]
+
+
+class Alias(object):
+    "adds aliases for the function"
+    def __init__(self, *aliases):
+        self.aliases = aliases
+    def __call__(self, func):
+        for alias in self.aliases:
+            Command.subcommands[alias] = func
+        return func
 
 class Command(object):
     "adds __annotations__ to decorated function, and adds func to Command.subcommands"
@@ -395,18 +429,20 @@ def usage(func, param_line_args):
     vararg = [vararg] if vararg else []
     keywordarg = [keywordarg] if keywordarg else []
     defaults = list(defaults) if defaults else []
-    #if not params:
-    #    raise ScriptionError("No parameters -- what's the point?")
     annotations = getattr(func, '__annotations__', {})
     indices = {}
     max_pos = 0
     positional = []
+    to_be_removed = []
     for i, name in enumerate(params + vararg + keywordarg):
-        spec = annotations.get(name, '')
-        help, kind, abbrev, type, choices, metavar = Spec(spec)
-        if name in keywordarg:
-            kind = 'keyword'
-        if kind == 'required' and name not in vararg + keywordarg:
+        spec = annotations.get(name, None)
+        if spec is None:
+            raise ScriptionError('%s not annotated' % name)
+        help, kind, abbrev, type, choices, usage_name, remove = Spec(spec)
+        if name in vararg + keywordarg:
+            if kind is None:
+                kind = 'option'
+        elif kind == 'required':
             max_pos += 1
             positional.append(empty)
         elif kind == 'flag':
@@ -415,11 +451,19 @@ def usage(func, param_line_args):
                 abbrev = name[0]
         elif kind == 'option':
             positional.append(None)
-        elif kind == 'keyword':
-            pass
+            if not abbrev:
+                abbrev = name[0]
+        elif kind == 'multi':
+            positional.append(tuple())
+            if not abbrev:
+                abbrev = name[0]
+        else:
+            raise ValueError('unknown kind: %r' % kind)
         if abbrev in annotations:
             raise ScriptionError('duplicate abbreviations: %r' % abbrev, ' '.join(param_line_args))
-        spec = Spec(help, kind, abbrev, type, choices, metavar)
+        if usage_name is None:
+            usage_name = name.upper()
+        spec = Spec(help, kind, abbrev, type, choices, usage_name, remove)
         annotations[i] = spec
         annotations[name] = spec
         indices[name] = i
@@ -450,17 +494,20 @@ def usage(func, param_line_args):
     program = param_line_args[0]
     print_params = []
     for param in params:
+        example = annotations[param].usage_name
         if annotations[param].kind == 'flag':
-            print_params.append('--' + param)
+            print_params.append('--%s' % param)
         elif annotations[param].kind == 'option':
-            print_params.append('--' + param + ' ...')
+            print_params.append('--%s %s' % (param, example))
+        elif annotations[param].kind == 'multi':
+            print_params.append('--%s %s [--%s ...]' % (param, example, param))
         else:
-            print_params.append(param)
+            print_params.append(example)
     usage = ["usage:", program] + print_params
     if vararg:
         usage.append("[%s [%s [...]]]" % (vararg[0], vararg[0]))
     if keywordarg:
-        usage.append("[%s=value [%s=value [...]]]" % (keywordarg[0], keywordarg[0]))
+        usage.append("[name1=value1 [name2=value2 [...]]]")
     usage = ['', ' '.join(usage), '']
     if func.__doc__:
         usage.extend(['    ' + func.__doc__.strip(), ''])
@@ -469,19 +516,19 @@ def usage(func, param_line_args):
         if posi is empty:
             posi = ''
         else:
-            posi = 'default: ' + repr(posi)
+            posi = '[default: ' + repr(posi) + ']'
         annote = annotations[name]
         choices = ''
         if annote.choices:
             choices = '[ %s ]' % ' | '.join(annote.choices)
         usage.append('    %-15s %s %s %s' % (
-            annote.metavar or name,
+            annote.usage_name,
             annote.help,
             posi,
             choices,
             ))
     for name in (vararg + keywordarg):
-        usage.append('    %-15s %s' % (annotations[name].metavar or name, annotations[name].help))
+        usage.append('    %-15s %s' % (name, annotations[name].help))
 
     func.__usage__ = '\n'.join(usage)
     args = []
@@ -489,22 +536,26 @@ def usage(func, param_line_args):
     pos = 0
     print_help = False
     value = None
-    errors = []
     rest = []
     doubledash = False
-    for item in param_line_args[1:] + [None]:
-        # required arguments /should/ be kept together
-        # once an option is found all text until the next option/flag/variable
-        # is part of that option
+    for offset, item in enumerate(param_line_args[1:] + [None]):
+        offset += 1
         original_item = item
         if value is not None:
             if item is None or item.startswith('-') or '=' in item:
-                value = annote.type(value.strip())
-                positional[index] = value
-                value = None
+                raise ScriptionError('%s has no value' % last_item)
+            to_be_removed.append(offset)
+            value.append(item)
+            if value[0][0] == '"':
+                if value[-1][-1] != '"':
+                    continue
+            value = annote.type(' '.join(value).strip('"'))
+            if isinstance(positional[index], tuple):
+                positional[index] += (value,)
             else:
-                value += ' ' + item
-                continue
+                positional[index] = value
+            value = None
+            continue
         if item is None:
             break
         if doubledash:
@@ -514,34 +565,50 @@ def usage(func, param_line_args):
             doubledash = True
             continue
         if item.startswith('-'):
-            if item == '--help':
+            if item.lower() == '--help':
                 print_help = True
                 continue
             item = item.lstrip('-')
             value = True
-            if item.startswith('no-') and '=' not in item:
+            if item.lower().startswith('no-') and '=' not in item:
                 value = False
                 item = item[3:]
             elif '=' in item:
                 item, value = item.split('=', 1)
+            item = item.replace('-','_')
             if item not in annotations:
-                if item in Script.settings:
+                if item in Script.settings or item in ('SCRIPTION_DEBUG', ):
                     Script.settings[item] = value
+                    value = None
+                    continue
                 else:
                     raise ScriptionError('%s not valid' % original_item, ' '.join(param_line_args))
             index = indices[item]
             annote = annotations[item]
-            if annote.kind == 'option' and value in (True, False):
-                value = ''
+            if annote.remove:
+                to_be_removed.append(offset)
+            if annote.kind in ('multi', 'option'):
+                if value in (True, False):
+                    value = []
+                    last_item = item
+                elif value[0] == '"':
+                    value = [value]
+                else:
+                    value = annote.type(value)
+                    if annote.kind == 'option':
+                        positional[index] = value
+                    else:
+                        positional[index] += (value,)
+                    value = None
             elif annote.kind == 'flag':
                 value = annote.type(value)
                 positional[index] = value
                 value = None
         elif '=' in item:
             item, value = item.split('=')
+            item = item.replace('-','_')
             if item in params:
-                errors.append('%s must be specified as a %s' % (item, annotations[item].kind))
-                continue
+                raise ScriptionError('%s must be specified as a %s' % (item, annotations[item].kind))
             item, value = keywordarg_type(item, value)
             if not isinstance(item, str):
                 raise ScriptionError('keyword names must be strings', ' '.join(param_line_args))
@@ -552,8 +619,7 @@ def usage(func, param_line_args):
                 annote = annotations[pos]
                 # check for choices membership before transforming into a type
                 if annote.choices and item not in annote.choices:
-                    errors.append('%r not in [ %s ]' % (item, ' | '.join(annote.choices)))
-                    continue
+                    raise ScriptionError('%r not in [ %s ]' % (item, ' | '.join(annote.choices)))
                 item = annote.type(item)
                 positional[pos] = item
                 pos += 1
@@ -562,51 +628,51 @@ def usage(func, param_line_args):
                 args.append(item)
     exc = None
     if args and rest:
-        errors.append('-- should be used to separate %s arguments from the rest' % program)
+        raise ScriptionError('-- should be used to separate %s arguments from the rest' % program)
     elif rest:
         args = rest
-    if errors:
-        print '\n' + '\n'.join(errors) #+ '\n\n'
-        raise ScriptionError('Invalid command line', command_line=' '.join(param_line_args))
     if print_help:
         print func.__usage__
         sys.exit()
     if not all([p is not empty for p in positional]):
-        raise ScriptionError('Invalid command line', command_line=' '.join(param_line_args))
-    if (args and not vararg
-    or  kwargs and not keywordarg
-    or  vararg and annotations[vararg[0]].kind == 'required' and not args
-    ):
-        raise ScriptionError('Invalid command line', command_line=' '.join(param_line_args))
+        raise ScriptionError('\n01 - Invalid command line:  %s' % ' '.join(param_line_args))
+    if args and not vararg:
+        raise ScriptionError("\n02 - don't know what to do with %s" % ', '.join(args))
+    elif kwargs and not keywordarg:
+        raise ScriptionError("\n03 - don't know what to do with %s" % ', '.join(['%s=%s' % (k, v) for k, v in kwargs.items()]))
+    elif vararg and annotations[vararg[0]].kind == 'required' and not args:
+        raise ScriptionError('\n04 - %s values are required\n' % vararg[0])
+    # remove any command line args that shouldn't be passed on
+    sys.argv[:] = [arg for (i, arg) in enumerate(sys.argv) if i not in to_be_removed]
     return tuple(positional + args), kwargs
 
-def Run(logger=None):
+def Run():
     "parses command-line and compares with either func or, if None, Script.command"
     module = None
+    debug = Script.settings.get('SCRIPTION_DEBUG')
     try:
         prog_name = Path(sys.argv[0]).filename
-        if logger:
-            logger.openlog(str(prog_name.filename), logger.LOG_PID)
+        if debug:
+            print(prog_name.filename)
         if Script.command and Command.subcommands:
             raise ScriptionError("scription does not support both Script and Command in the same file")
         if Script.command is None and not Command.subcommands:
             raise ScriptionError("either Script or Command must be specified")
         if Command.subcommands:
-            func = Command.subcommands.get(prog_name, None)
+            func_name = sys.argv[1:2]
+            if not func_name:
+                func = None
+            else:
+                func = Command.subcommands.get(func_name[0])
             if func is not None:
                 module = func.func_globals
-                prog_name = sys.argv[0]
-                param_line = [prog_name] + sys.argv[1:]
+                prog_name = func_name[0]
+                param_line = [prog_name] + sys.argv[2:]
             else:
-                func_name = sys.argv[1:2]
-                if not func_name:
-                    func = None
-                else:
-                    func = Command.subcommands.get(func_name[0])
-                if func and func is not None:
+                func = Command.subcommands.get(prog_name, None)
+                if func is not None:
                     module = func.func_globals
-                    prog_name = ' '.join(sys.argv[:2])
-                    param_line = [prog_name] + sys.argv[2:]
+                    param_line = [prog_name] + sys.argv[1:]
                 else:
                     for name, func in sorted(Command.subcommands.items()):
                         try:
@@ -622,15 +688,16 @@ def Run(logger=None):
         args, kwargs = usage(func, param_line)
         func.func_globals.update(Script.settings)
         result = func(*args, **kwargs)
-        if logger:
-            logger.syslog('done')
-            logger.closelog()
         return result
     except Exception:
-        if logger:
-            result = log_exception()
-            if module:
-                module['exception_lines'] = result
+        exc = sys.exc_info()[1]
+        if debug:
+            print exc
+        result = log_exception()
+        if module:
+            module['exception_lines'] = result
+        if isinstance(exc, ScriptionError):
+            raise SystemExit(str(exc))
         raise
 
 def InputFile(arg):
